/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.catalina.startup;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.catalina.Globals;
import org.apache.catalina.security.SecurityClassLoad;
import org.apache.catalina.startup.ClassLoaderFactory.Repository;
import org.apache.catalina.startup.ClassLoaderFactory.RepositoryType;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;
import org.chenchen.test.LogConstantForCode;

/**
 * Bootstrap loader for Catalina.  This application constructs a class loader
 * for use in loading the Catalina internal classes (by accumulating all of the
 * JAR files found in the "server" directory under "catalina.home"), and
 * starts the regular execution of the container.  The purpose of this
 * roundabout approach is to keep the Catalina internal classes (and any
 * other classes they depend on, such as an XML parser) out of the system
 * class path and therefore not visible to application level classes.
 * Bootstrap为了Servlet容器而先加载.为了装入Catalina内部类(通过在“Catalina .home”
 * 下的“server”目录中找到的所有JAR文件)， 这个应用程序构造了一个类加载器。并开始执行
 * 容器的常规操作。这种加载方式是为了将Catalina内部类(以及它们所依赖的任何其他类，如XML
 * 解析器)排除在系统类路径之外。所以对application级别的类是不可见的。
 *
 * 启动器类。tomcat启动的入口类。请先关注static块儿的内容。
 *
 * @author Craig R. McClanahan
 * @author Remy Maucherat
 * @translator chenchen6(chenmudu@gmail.com/chenchen6@tuhu.cn)
 */
public final class Bootstrap {

    private static final Log log = LogFactory.getLog(Bootstrap.class);

    /**
     * Daemon object used by main.
     */
    private static final Object daemonLock = new Object();
    /**
     * 守护进程在init的时候被自身设置.设置了一系列的ClassLoader。
     */
    private static volatile Bootstrap daemon = null;

    private static final File catalinaBaseFile;
    private static final File catalinaHomeFile;

    private static final Pattern PATH_PATTERN = Pattern.compile("(\".*?\")|(([^,])*)");

    /**

     * 1.静态代码块儿在类中的加载顺序。详见深入理解JVM。
     *  1.1 获取bin目录的绝对路径。
     *  1.2.设置安装目录的路径。
     *  1.3.设置工作目录的路径。
     * 也就是启动参数中的参数设置：
     *  -Dcatalina.home=catalina-home   //安装目录。
     *  -Dcatalina.base=catalina-home   //工作目录。
     *  从环境变量找,找不到后就去设置即可。
     *
     */
    static {
        log.info("Bootstrap static {} start!");
        // Will always be non-null
        /**
         * C:\Users\chenc\Downloads\MySelf\Tomcat8-Source-Read ($CATALINA_HOME\bin：用户目录的绝对路径。)
         */
        String userDir = System.getProperty("user.dir");

        // Home first
        /**
         * catalina-home (Tomcat的根目录：当前项目的相对路径)
         */
        String home = System.getProperty(Globals.CATALINA_HOME_PROP);
        File homeFile = null;

        //获取的是tomcat的安装目录.即：磁盘下catalina-home下绝对路径下的所有的文件夹和文件封装成File。
        if (home != null) {
            File f = new File(home);
            try {
                homeFile = f.getCanonicalFile();
            } catch (IOException ioe) {
                homeFile = f.getAbsoluteFile();
            }
        }

        if (homeFile == null) {
            // First fall-back. See if current directory is a bin directory
            // in a normal Tomcat install
            //正常安装下查看是不是bin目录。
            File bootstrapJar = new File(userDir, "bootstrap.jar");

            if (bootstrapJar.exists()) {
                File f = new File(userDir, "..");
                try {
                    homeFile = f.getCanonicalFile();
                } catch (IOException ioe) {
                    homeFile = f.getAbsoluteFile();
                }
            }
        }

        if (homeFile == null) {
            // Second fall-back. Use current directory
            File f = new File(userDir);
            try {
                homeFile = f.getCanonicalFile();
            } catch (IOException ioe) {
                homeFile = f.getAbsoluteFile();
            }
        }
        //工作目录和安装目录默认是一致的。可以设为不一致。
        //环境变量设置(安装目录)。 设置catalinaHomeFile = 当前bin文件夹绝对路径。
        catalinaHomeFile = homeFile;
        System.setProperty(
                Globals.CATALINA_HOME_PROP, catalinaHomeFile.getPath());

        // Then base
        String base = System.getProperty(Globals.CATALINA_BASE_PROP);
        if (base == null) { //设置工作目录。
            catalinaBaseFile = catalinaHomeFile;
        } else {
            File baseFile = new File(base);
            try {
                baseFile = baseFile.getCanonicalFile();
            } catch (IOException ioe) {
                baseFile = baseFile.getAbsoluteFile();
            }
            catalinaBaseFile = baseFile;
        }
        System.setProperty(
                Globals.CATALINA_BASE_PROP, catalinaBaseFile.getPath());
        log.info("Bootstrap static {} end!");
    }

    // -------------------------------------------------------------- Variables


    /**
     * Daemon reference.
     */
    private Object catalinaDaemon = null;
    /**
     * tomcat自己定义的三个类加载器(URLClassLoader子类).为了打破双亲委派模型。
     *  JVM类加载器结构图:(双亲委派模型)
     *          Bootstrap ClassLoader    --
     *                  |                  |
     *           Ext ClassLoader           |----> JVM内部的三个类加载器。
     *                  |                  |
     *           System ClassLoader      --
     *                  |
     *           Common ClassLoader     --->Tomcat自定义根加载器(tomcat公共类加载器)
     *    -------------------------------
     *          |               |
     *  CatalinaLoader     SharedLoader-->所有的Web服务器的类加载器。(web向从这里找。)
     *                          |
     *                        WebApps
     *                          |
     *                   -------------
     *                      |       |
     *                   WebApps  WebApps
     *                     |
     *           (WEB-INFO/classes,lib)
     *  ()
     */
    ClassLoader commonLoader = null;
    ClassLoader catalinaLoader = null;
    ClassLoader sharedLoader = null;


    // -------------------------------------------------------- Private Methods

    /**
     * 初始化各种类加载器。
     */
    private void initClassLoaders() {
        try {
            //置空  目的：打破双亲委派模型机制。因为浪费性能。
            commonLoader = createClassLoader("common", null);
            if (commonLoader == null) {
                // no config file, default to this loader - we might be in a 'single' env.
                /**
                 * Connected to the target VM, address:127.0.0.1:1083 tranpsort:'socket'
                 * 从Console中的日志可以看出。网络通讯有关。
                 * Bootstrap.getClassLoader = URLClassLoader.
                 */
                commonLoader = this.getClass().getClassLoader();
            }
            log.info("org.apache.catalina.startup.Bootstrap.initClassLoaders commonLoader 类型为:"+ commonLoader.getClass().getName());
            //指向了commonLoader
            catalinaLoader = createClassLoader("server", commonLoader);//(打一个断点观察catalinaLoader的值)
            log.info("org.apache.catalina.startup.Bootstrap.initClassLoaders catalinaLoader 类型为:"+ catalinaLoader.getClass().getName());
            //指向了commonLoader
            sharedLoader = createClassLoader("shared", commonLoader);//(打断点观测sharedLoader的值。)
            log.info("org.apache.catalina.startup.Bootstrap.initClassLoaders commonLoader 类型为:"+ commonLoader.getClass().getName());

        } catch (Throwable t) {
            handleThrowable(t);
            log.error("Class loader creation threw exception", t);
            System.exit(1);
        }
    }

    /**
     * 创建类加载器
     * @param name
     * @param parent
     * @return
     * @throws Exception
     */
    private ClassLoader createClassLoader(String name, ClassLoader parent)
            throws Exception {

        String value = CatalinaProperties.getProperty(name + ".loader");
        if ((value == null) || (value.equals("")))
            return parent;

        value = replace(value);

        List<Repository> repositories = new ArrayList<>();

        String[] repositoryPaths = getPaths(value);

        for (String repository : repositoryPaths) {
            // Check for a JAR URL repository
            try {
                @SuppressWarnings("unused")
                URL url = new URL(repository);
                repositories.add(new Repository(repository, RepositoryType.URL));
                continue;
            } catch (MalformedURLException e) {
                // Ignore
            }

            // Local repository
            if (repository.endsWith("*.jar")) {
                repository = repository.substring
                        (0, repository.length() - "*.jar".length());
                repositories.add(new Repository(repository, RepositoryType.GLOB));
            } else if (repository.endsWith(".jar")) {
                repositories.add(new Repository(repository, RepositoryType.JAR));
            } else {
                repositories.add(new Repository(repository, RepositoryType.DIR));
            }
        }

        return ClassLoaderFactory.createClassLoader(repositories, parent);
    }


    /**
     * System property replacement in the given string.
     *
     * @param str The original string
     * @return the modified string
     */
    protected String replace(String str) {
        // Implementation is copied from ClassLoaderLogManager.replace(),
        // but added special processing for catalina.home and catalina.base.
        String result = str;
        int pos_start = str.indexOf("${");
        if (pos_start >= 0) {
            StringBuilder builder = new StringBuilder();
            int pos_end = -1;
            while (pos_start >= 0) {
                builder.append(str, pos_end + 1, pos_start);
                pos_end = str.indexOf('}', pos_start + 2);
                if (pos_end < 0) {
                    pos_end = pos_start - 1;
                    break;
                }
                String propName = str.substring(pos_start + 2, pos_end);
                String replacement;
                if (propName.length() == 0) {
                    replacement = null;
                } else if (Globals.CATALINA_HOME_PROP.equals(propName)) {
                    replacement = getCatalinaHome();
                } else if (Globals.CATALINA_BASE_PROP.equals(propName)) {
                    replacement = getCatalinaBase();
                } else {
                    replacement = System.getProperty(propName);
                }
                if (replacement != null) {
                    builder.append(replacement);
                } else {
                    builder.append(str, pos_start, pos_end + 1);
                }
                pos_start = str.indexOf("${", pos_end + 1);
            }
            builder.append(str, pos_end + 1, str.length());
            result = builder.toString();
        }
        return result;
    }


    /**
     * Initialize daemon.
     * @throws Exception Fatal initialization error
     *
     * 1. 初始化类加载器。
     * 2. 实例化Catalina对象。 daemon的设置。
     * 3. Catalina去init Server对象。
     */
    public void init() throws Exception {
        //初始化类加载器。
        initClassLoaders();//(断点在这儿打一个。)
        Thread.currentThread().setContextClassLoader(catalinaLoader);

        SecurityClassLoad.securityClassLoad(catalinaLoader);

        // Load our startup class and call its process() method
        if (log.isDebugEnabled())
            log.debug("Loading startup class");
        //反射去加载Catalina类并进行实例化。
        Class<?> startupClass = catalinaLoader.loadClass("org.apache.catalina.startup.Catalina");
        Object startupInstance = startupClass.getConstructor().newInstance();

        // Set the shared extensions class loader
        if (log.isDebugEnabled())
            log.debug("Setting startup class properties");
        String methodName = "setParentClassLoader";
        Class<?> paramTypes[] = new Class[1];
        paramTypes[0] = Class.forName("java.lang.ClassLoader");
        Object paramValues[] = new Object[1];
        paramValues[0] = sharedLoader;
        /**
         * 反射执行对应setParentClassLoader方法。
         * 详见{@link Catalina#setParentClassLoader(ClassLoader)}
         * (打断点观测下获取的方法的名称和参数 以及真正执行的方法)
         */
        Method method =
                startupInstance.getClass().getMethod(methodName, paramTypes);
        method.invoke(startupInstance, paramValues);
        //将Catalina设置会守护进程。
        catalinaDaemon = startupInstance;
    }


    /**
     * Load daemon.
     * load的时候 这时候有参数了。
     */
    private void load(String[] arguments) throws Exception {

        // Call the load() method
        String methodName = "load";
        Object param[];
        Class<?> paramTypes[];
        if (arguments==null || arguments.length==0) {
            paramTypes = null;
            param = null;
        } else {
            paramTypes = new Class[1];
            paramTypes[0] = arguments.getClass();
            param = new Object[1];
            param[0] = arguments;
        }
        Method method =
                catalinaDaemon.getClass().getMethod(methodName, paramTypes);
        if (log.isDebugEnabled()) {
            log.debug("Calling startup class " + method);
        }
        //反射调用Catalina。
        method.invoke(catalinaDaemon, param);
    }


    /**
     * getServer() for configtest
     */
    private Object getServer() throws Exception {

        String methodName = "getServer";
        Method method = catalinaDaemon.getClass().getMethod(methodName);
        return method.invoke(catalinaDaemon);
    }


    // ----------------------------------------------------------- Main Program


    /**
     * Load the Catalina daemon.
     * @param arguments Initialization arguments
     * @throws Exception Fatal initialization error
     */
    public void init(String[] arguments) throws Exception {

        init();
        load(arguments);
    }


    /**
     * Start the Catalina daemon.
     * 反射去调用Catalina的start方法。{@link Catalina#start()}
     * @throws Exception Fatal start error
     */
    public void start() throws Exception {
        if (catalinaDaemon == null) {
            init();
        }

        Method method = catalinaDaemon.getClass().getMethod("start", (Class [])null);
        /**
         * 反射去调用Catalina的start方法。{@link Catalina#start()}
         */
        method.invoke(catalinaDaemon, (Object [])null);
    }


    /**
     * Stop the Catalina Daemon.
     * @throws Exception Fatal stop error
     */
    public void stop() throws Exception {
        Method method = catalinaDaemon.getClass().getMethod("stop", (Class []) null);
        method.invoke(catalinaDaemon, (Object []) null);
    }


    /**
     * Stop the standalone server.
     * @throws Exception Fatal stop error
     */
    public void stopServer() throws Exception {

        Method method =
                catalinaDaemon.getClass().getMethod("stopServer", (Class []) null);
        method.invoke(catalinaDaemon, (Object []) null);
    }


    /**
     * Stop the standalone server.
     * @param arguments Command line arguments
     * @throws Exception Fatal stop error
     */
    public void stopServer(String[] arguments) throws Exception {

        Object param[];
        Class<?> paramTypes[];
        if (arguments == null || arguments.length == 0) {
            paramTypes = null;
            param = null;
        } else {
            paramTypes = new Class[1];
            paramTypes[0] = arguments.getClass();
            param = new Object[1];
            param[0] = arguments;
        }
        Method method =
                catalinaDaemon.getClass().getMethod("stopServer", paramTypes);
        method.invoke(catalinaDaemon, param);
    }


    /**
     * Set flag.
     * @param await <code>true</code> if the daemon should block
     * @throws Exception Reflection error
     */
    public void setAwait(boolean await)
            throws Exception {

        Class<?> paramTypes[] = new Class[1];
        paramTypes[0] = Boolean.TYPE;
        Object paramValues[] = new Object[1];
        paramValues[0] = Boolean.valueOf(await);
        Method method =
                catalinaDaemon.getClass().getMethod("setAwait", paramTypes);
        method.invoke(catalinaDaemon, paramValues);
    }

    public boolean getAwait() throws Exception {
        Class<?> paramTypes[] = new Class[0];
        Object paramValues[] = new Object[0];
        Method method =
                catalinaDaemon.getClass().getMethod("getAwait", paramTypes);
        Boolean b=(Boolean)method.invoke(catalinaDaemon, paramValues);
        return b.booleanValue();
    }


    /**
     * Destroy the Catalina Daemon.
     */
    public void destroy() {

        // FIXME

    }


    /**
     * 2.运行的基础。(instance, init, load, start)
     * 2.1 创建Bootstrap(实例化)。
     * 2.2 检测运行命令。
     * 2.3 执行对应操作。
     * 2.4 Bootstrap初始化自身。(init自身。)
     * 2.5 加载(load)Catalina。
     *          2.5.1. Catalina初始化(init)Server
     *              2.5.1.1 Server去初始化(init)了Service。
     *                  2.5.1.1.1 Service去初始化(init)了Engine。
     *                  2.5.1.1.2 Service去初始化(init)了Executor。
     *                  2.5.1.1.3 Service去初始化(init)了Connector。
     *                      2.5.1.1.3.1 Connector去初始化了ProtocolHandler.
     *                          2.5.1.1.3.1.1 ProtocolHandler 初始化了EndPonit.
     *          2.5.2  Catalina(开启) start了Server。
     *              2.5.2.1 Server去开启(start)了Service。
     *                  2.5.1.1.1 Service去开启(start)了Engine。
     *                  2.5.1.1.2 Service去开启(start)了Executor。
     *                  2.5.1.1.3 Service去开启(start)了Connector。
     *                      2.5.1.1.3.1 Connector去开启(start)了ProtocolHandler.
     *                          2.5.1.1.3.1.1 ProtocolHandler 开启(start)了EndPonit.
     *
     * Main method and entry point when starting Tomcat via the provided
     * scripts.
     * 通过提供的脚本去执行入口方法.这是启动tomcat的入口。
     * @param args Command line arguments to be processed
     */
    public static void main(String args[]) {
        log.info("current your machine time is : " + LogConstantForCode.CURRENT_LOCAL_DATE);
        //断点可以在这儿打一个。
        synchronized (daemonLock) {
            if (daemon == null) { //tomcat没用启动。
                // Don't set daemon until init() has completed
                //在init()方法完成前不要设置守护线程进程
                Bootstrap bootstrap = new Bootstrap();
                try {
                    bootstrap.init();
                } catch (Throwable t) {
                    handleThrowable(t);
                    t.printStackTrace();
                    return;
                }
                daemon = bootstrap;//设置守护即引用Bootstrap。
            } else {
                // When running as a service the call to stop will be on a new
                // thread so make sure the correct class loader is used to
                // prevent a range of class not found exceptions.
                Thread.currentThread().setContextClassLoader(daemon.catalinaLoader);
            }
        }
        /**
         * 此时daemon已经变成Catalina了。执行load方法其实是在执行Catalina的load方法。
         *   详见{@link Catalina#load}
         */
        try {
            String command = "start";
            if (args.length > 0) {
                command = args[args.length - 1];
            }

            if (command.equals("startd")) {
                args[args.length - 1] = "start";

                daemon.load(args); //
                daemon.start();
            } else if (command.equals("stopd")) {
                args[args.length - 1] = "stop";
                daemon.stop();
            } else if (command.equals("start")) {
                /**
                 * daemon 变成了Catalina。去初始化了Catalina。
                 * 并将Catalina.await设置为true.{@link Catalina#await()}
                 * 目的：在关闭窗口阻塞住监听关闭。将线程阻塞住。
                 * {@link Catalina#load()}
                 * Catalina.setAwait, Catalina.load, Catalina.start()。。。。。
                 */
                daemon.setAwait(true);
                daemon.load(args);
                /**
                 * 容器启动的入口。
                 * {@link Bootstrap#start()}
                 */
                daemon.start();
                if (null == daemon.getServer()) {
                    System.exit(1);
                }
            } else if (command.equals("stop")) {
                daemon.stopServer(args);
            } else if (command.equals("configtest")) {
                daemon.load(args);
                if (null == daemon.getServer()) {
                    System.exit(1);
                }
                System.exit(0);
            } else {
                log.warn("Bootstrap: command \"" + command + "\" does not exist.");
            }
        } catch (Throwable t) {
            // Unwrap the Exception for clearer error reporting
            if (t instanceof InvocationTargetException &&
                    t.getCause() != null) {
                t = t.getCause();
            }
            handleThrowable(t);
            t.printStackTrace();
            System.exit(1);
        }
    }


    /**
     * Obtain the name of configured home (binary) directory. Note that home and
     * base may be the same (and are by default).
     * @return the catalina home
     */
    public static String getCatalinaHome() {
        return catalinaHomeFile.getPath();
    }


    /**
     * Obtain the name of the configured base (instance) directory. Note that
     * home and base may be the same (and are by default). If this is not set
     * the value returned by {@link #getCatalinaHome()} will be used.
     * @return the catalina base
     */
    public static String getCatalinaBase() {
        return catalinaBaseFile.getPath();
    }


    /**
     * Obtain the configured home (binary) directory. Note that home and
     * base may be the same (and are by default).
     * @return the catalina home as a file
     */
    public static File getCatalinaHomeFile() {
        return catalinaHomeFile;
    }


    /**
     * Obtain the configured base (instance) directory. Note that
     * home and base may be the same (and are by default). If this is not set
     * the value returned by {@link #getCatalinaHomeFile()} will be used.
     * @return the catalina base as a file
     */
    public static File getCatalinaBaseFile() {
        return catalinaBaseFile;
    }


    // Copied from ExceptionUtils since that class is not visible during start
    private static void handleThrowable(Throwable t) {
        if (t instanceof ThreadDeath) {
            throw (ThreadDeath) t;
        }
        if (t instanceof VirtualMachineError) {
            throw (VirtualMachineError) t;
        }
        // All other instances of Throwable will be silently swallowed
    }


    // Protected for unit testing
    protected static String[] getPaths(String value) {

        List<String> result = new ArrayList<>();
        Matcher matcher = PATH_PATTERN.matcher(value);

        while (matcher.find()) {
            String path = value.substring(matcher.start(), matcher.end());

            path = path.trim();
            if (path.length() == 0) {
                continue;
            }

            char first = path.charAt(0);
            char last = path.charAt(path.length() - 1);

            if (first == '"' && last == '"' && path.length() > 1) {
                path = path.substring(1, path.length() - 1);
                path = path.trim();
                if (path.length() == 0) {
                    continue;
                }
            } else if (path.contains("\"")) {
                // Unbalanced quotes
                // Too early to use standard i18n support. The class path hasn't
                // been configured.
                throw new IllegalArgumentException(
                        "The double quote [\"] character only be used to quote paths. It must " +
                                "not appear in a path. This loader path is not valid: [" + value + "]");
            } else {
                // Not quoted - NO-OP
            }

            result.add(path);
        }
        return result.toArray(new String[result.size()]);
    }
}
